import asyncio
from dataclasses import dataclass
from enum import Enum
import json
from urllib.parse import quote

from aiohttp import ClientSession

from raphson_music_client.lyrics import LyricsLine, PlainLyrics, TimeSyncedLyrics


class AudioFormat(Enum):
    WEBM_OPUS_HIGH = 'webm_opus_high'
    WEBM_OPUS_LOW = 'webm_opus_low'
    MP4_AAC = 'mp4_aac'
    MP3_WITH_METADATA = 'mp3_with_metadata'


@dataclass
class DownloadedTrack:
    track: "Track | None"
    audio: bytes
    image: bytes
    lyrics_json: str

    @property
    def lyrics(self) -> TimeSyncedLyrics | PlainLyrics | None:
        lyrics_dict = json.loads(self.lyrics_json)
        if lyrics_dict["type"] == "none":
            return None
        elif lyrics_dict["type"] == "plain":
            return PlainLyrics(lyrics_dict["source"], lyrics_dict["text"])
        elif lyrics_dict["type"] == "synced":
            text = [
                LyricsLine(line["start_time"], line["text"])
                for line in lyrics_dict["text"]
            ]
            return TimeSyncedLyrics(lyrics_dict["source"], text)
        else:
            raise ValueError()

    @property
    def lyrics_text(self) -> str | None:
        lyrics = self.lyrics
        if not lyrics:
            return None
        if isinstance(lyrics, TimeSyncedLyrics):
            lyrics = lyrics.to_plain()
        return lyrics.text


@dataclass
class Track:
    path: str
    display: str
    mtime: int
    duration: int
    title: str | None
    album: str | None
    album_artist: str | None
    year: int | None
    artists: list[str]

    @property
    def playlist(self):
        return self.path[self.path.index('/'):]


@dataclass
class DownloadableTrack(Track):
    _session: ClientSession

    async def get_audio(self, audio_format: AudioFormat) -> bytes:
        response = await self._session.get(
            "/track/" + quote(self.path) + "/audio?type=" + audio_format.value
        )
        return await response.content.read()

    async def get_cover_image(self) -> bytes:
        response = await self._session.get(
            "/track/" + quote(self.path) + "/cover?quality=high"
        )
        return await response.content.read()

    async def get_lyrics_json(self) -> str:
        response = await self._session.get(
            "/track/" + quote(self.path) + "/lyrics"
        )
        return await response.text()

    async def download(self, audio_format: AudioFormat = AudioFormat.WEBM_OPUS_HIGH) -> DownloadedTrack:
        audio, image, lyrics_json = await asyncio.gather(
            self.get_audio(audio_format),
            self.get_cover_image(),
            self.get_lyrics_json(),
        )
        return DownloadedTrack(self, audio, image, lyrics_json)

    async def download_mp3(self):
        return await self.get_audio(AudioFormat.MP3_WITH_METADATA)



from dataclasses import dataclass
from aiohttp import ClientSession, ClientTimeout

from raphson_music_client.track import Track


@dataclass
class PlayerState:
    all_playlists: list[str]
    enabled_playlists: list[str]
    has_media: bool
    is_playing: bool
    position: int
    duration: int
    volume: float
    track: Track | None


class HeadlessPlayerClient:
    _session: ClientSession

    def __init__(self):
        self._session = None  # pyright: ignore[reportAttributeAccessIssue]

    async def setup(self, *, base_url: str) -> None:
        self._session = ClientSession(
            base_url=base_url,
            timeout=ClientTimeout(connect=5, total=60),
            raise_for_status=True,
        )

    async def state(self) -> PlayerState:
        response = await self._session.get("/state")
        json = await response.json()
        currently_playing = json["currently_playing"]
        if currently_playing:
            track = Track(
                currently_playing["path"],
                currently_playing["path"], # TODO proper display string
                0,  # TODO mtime
                currently_playing["duration"],
                currently_playing["title"],
                currently_playing["album"],
                currently_playing["album_artist"],
                currently_playing["year"],
                currently_playing["artists"],
            )
        else:
            track = None

        return PlayerState(
            json["playlists"]["all"],
            json["playlists"]["enabled"],
            json["player"]["has_media"],
            json["player"]["is_playing"],
            json["player"]["position"],
            json["player"]["duration"],
            json["player"]["volume"] / 100,
            track,
        )

    async def list_tracks(self, playlist: str) -> list[Track]:
        response = await self._session.get("/list_tracks", params={"playlist": playlist})
        response_json = await response.json()
        tracks: list[Track] = []
        for track_json in response_json:
            tracks.append(
                Track(
                    track_json["path"],
                    track_json["display"],
                    track_json["mtime"],
                    track_json["duration"],
                    track_json["title"],
                    track_json["album"],
                    track_json["album_artist"],
                    track_json["year"],
                    track_json["artists"],
                )
            )
        return tracks

    async def stop(self) -> None:
        await self._session.post('/stop')

    async def pause(self) -> None:
        await self._session.post('/pause')

    async def play(self) -> None:
        await self._session.post('/play')

    async def next(self) -> None:
        await self._session.post('/next')

    async def play_news(self) -> None:
        await self._session.post('/play_news')

    async def seek(self, position: int) -> None:
        await self._session.post('/seek', data=str(position))

    async def set_volume(self, volume: float) -> None:
        await self._session.post('/volume', data=str(int(volume * 100)))

    async def set_enabled_playlists(self, playlists: list[str]) -> None:
        await self._session.post('/playlists', json=playlists)

    async def enqueue(self, track: str | Track) -> None:
        if isinstance(track, Track):
            track = track.path
        await self._session.post('/enqueue', data=track)

    async def play_track(self, track: str | Track) -> None:
        if isinstance(track, Track):
            track = track.path
        await self._session.post('/play_track', data=track)



from dataclasses import dataclass


@dataclass
class LyricsLine:
    start_time: float
    text: str


@dataclass
class TimeSyncedLyrics():
    source: str
    text: list[LyricsLine]

    def to_plain(self) -> 'PlainLyrics':
        text = '\n'.join([line.text for line in self.text])
        return PlainLyrics(self.source, text)


@dataclass
class PlainLyrics():
    source: str
    text: str



from abc import ABC
import asyncio
import json
import uuid
from dataclasses import dataclass
from typing import Any
from urllib.parse import quote

from aiohttp import ClientTimeout, StreamReader
from aiohttp.client import ClientError, ClientResponse, ClientSession

from raphson_music_client.playlist import Playlist
from raphson_music_client.track import DownloadableTrack, DownloadedTrack, Track


class RaphsonMusicClient:
    player_id: str
    session: ClientSession
    cached_rapson_logo: bytes | None = None

    def __init__(self):
        self.player_id = str(uuid.uuid4())
        self.session = None  # pyright: ignore[reportAttributeAccessIssue]

    async def setup(self, *, base_url: str, user_agent: str, token: str) -> None:
        self.session = ClientSession(
            base_url=base_url,
            headers={"User-Agent": user_agent, "Authorization": "Bearer " + token},
            timeout=ClientTimeout(connect=5, total=60),
            raise_for_status=True,
        )

    async def close(self) -> None:
        if self.session:
            await self.session.close()

    def _track_from_json(self, json: dict[str, Any]) -> DownloadableTrack:
        assert self.session
        return DownloadableTrack(
            json["path"],
            json["display"],
            json["mtime"],
            json["duration"],
            json["title"],
            json["album"],
            json["album_artist"],
            json["year"],
            json["artists"],
            self.session,
        )

    async def choose_track(self, playlist: str) -> DownloadableTrack:
        response = await self.session.post(
            "/playlist/" + quote(playlist) + "/choose_track", json={}
        )
        json = await response.json()
        return self._track_from_json(json)

    async def get_track(self, path: str) -> DownloadableTrack:
        response = await self.session.get("/track/" + quote(path) + "/info")
        json = await response.json()
        return self._track_from_json(json)

    async def submit_now_playing(self, track_path: str, position: int, paused: bool) -> None:
        await self.session.post(
            "/activity/now_playing",
            json={
                "player_id": self.player_id,
                "track": track_path,
                "paused": paused,
                "position": position,
            },
        )

    async def submit_played(self, track_path: str, timestamp: int) -> None:
        await self.session.post(
            "/activity/played", json={"track": track_path, "timestamp": timestamp}
        )

    async def _get_news_audio(self) -> bytes:
        response = await self.session.get("/news/audio")
        return await response.content.read()

    async def get_news(self) -> DownloadedTrack:
        audio, image = await asyncio.gather(self._get_news_audio(), self.get_raphson_logo())
        return DownloadedTrack(None, audio, image, '{"type":"none"}')

    async def get_raphson_logo(self) -> bytes:
        if not self.cached_rapson_logo:
            response = await self.session.get("/static/img/raphson.png")
            self.cached_rapson_logo = await response.content.read()
        return self.cached_rapson_logo

    async def list_tracks_response(self, playlist: str) -> StreamReader:
        response = await self.session.get("/tracks/filter?playlist=" + quote(playlist))
        return response.content

    async def list_tracks(self, playlist: str) -> list[DownloadableTrack]:
        response = await self.session.get("/tracks/filter?playlist=" + quote(playlist))
        response_json = await response.json()
        return [self._track_from_json(track_json) for track_json in response_json["tracks"]]

    async def playlists(self) -> list[Playlist]:
        response = await self.session.get("/playlist/list")
        return [
            Playlist(playlist["name"], playlist["favorite"]) for playlist in await response.json()
        ]
